/* xmdserver.c */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>   

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h> 
#include <netdb.h>


/* for FreeBSD */
#ifndef TCP_NODELAY
#include <netinet/tcp.h>
#endif

#define CLI_MAX 10         /* $B:GBg%/%i%$%"%s%H?t(B($B:GBg(B32)*/
#define UNUSED (-1)        /*$B%=%1%C%H$,;H$o$l$F$$$J$$>uBV(B*/
#define DEATH '*'          /*$B%2!<%`%*!<%P!<>uBV(B*/
#define WIDTH 8            /*1$B9T$"$?$j$N$W$h$N8D?t(B*/
#define HEIGHT 11          /*$B2hLL>e$N9T?t(B($B$?$@$7!"(B0$B$^$G!K(B*/
#define INSTR_SIZE 10      /*$BL?Na8l$N:GBgD9(B*/
#define BEG_LINE   4       /*$B:G=i$KAw$k9T?t(B*/
#define ALLSTATE_LENGTH 100 /*ALLSTATE$BL?Na$H$7$FAw$kJ8;zNs$N:GBgD9(B*/
#define LIST_LENGTH 150    /*LIST$BL?Na$H$7$FAw$kJ8;zNs$N:GBgD9(B*/
#define PUYO_LENGTH 30     /*PUYO$BL?Na$H$7$FAw$kJ8;zNs$N:GBgD9(B*/
#define NICK 1   
#define START 2
#define REQUEST 3
#define MYSTATE 4
#define ERROR (-1)

#define SELECT_TIMEOUT 1

/*$BMp?t$rH/@8$5$;$k4X?t(Brandomi()$B$G;H$&%^%/%m(B*/
#define LAMBDA 843314861L
#define RANCONST 453816693L
#define RANDOM_INIT_VALUE 198313L
#define RAN2_32 4294967296.0

/*$B30ItJQ?t(B*/
char allstate[ALLSTATE_LENGTH] = "ALLSTATE";  /*ALLSTATE$BL?Na$H$7$FAw$kJ8;zNs(B*/
char list[LIST_LENGTH] = "LIST";          /*LIST$BL?Na$H$7$FAw$kJ8;zNs(B*/
char puyo[PUYO_LENGTH] = "PUYO:";
char mystate[30] = "MYSTATE:*:*;";


int main(int argc, char *argv[])
{
    int state;                /*$B3+;OA0$J$i(B0$B!"%2!<%`Cf$J$i(B1*/
    int max;                  /*$B8=:_$N%(%s%H%j?t!J<!$N6u$-%=%1%C%HHV9f!K(B*/
    int PORT;
    int newcomer;             /*$B?75,<u$1IU$1MQ%=%1%C%H%G%9%/%j%W%?(B*/
    int s[CLI_MAX + 1];       /* $B%/%i%$%"%s%H@\B3MQ%=%1%C%H%G%9%/%j%W%?(B*/
    int pd[CLI_MAX + 1];
    int touroku[CLI_MAX + 1];

    int len;                  /*caddr$B$N%5%$%:(B*/
    fd_set readfds;
    struct sockaddr_in saddr; /* $B%5!<%PB&(B($B<+J,(B)$B$N%"%I%l%9>pJs(B */
    struct sockaddr_in caddr; /* $B%/%i%$%"%s%HB&$N%"%I%l%9>pJs$rF~$l$k$H$3$m(B */
    char str[1024];           /* $B<u?.MQ%P%C%U%!(B*/
    int i, j;                 /* $B%k!<%W%+%&%s%?(B*/
    int msglen;               /*$B<u?.$7$?%a%C%;!<%8$ND9$5(B*/
    int n, r;                    /*select()$B$NJV5QCM(B*/
    int pd0, pd1, win_pd;
    int connect_;
    int sockflag = 1;
    struct timeval timeout = {SELECT_TIMEOUT, 0};
    time_t  now_time, pre_time;

    int  get_port(char *argv[]);
    int  analysis(char *str, int state);
    int  winner(int max);
    void broadcast(char *str, int *s, int max);
    void list_update(char *str, int pd);
    void nickname_erase(int pd);
    void allstate_update(char *str);
    void puyo_make(int pd);
    void win_make(char *str, int pd);
    int  nickname_search(char *str);

    if( argc ==1){
        PORT = 4170;
      }else if(argc != 3){
        fprintf(stderr, "xmdserver v2.0\nusage: xmdserver [-p port]\n");
        exit(1);
    }else if((PORT = get_port(argv)) == -1){
        fprintf(stderr, "xmdserver v2.0\nusage: xmdserver [-p port]\n");
        exit(1);
    }
    /*$BJQ?t$N=i4|2=(B*/
    state = 0;
    max = 0;

    for(i = 0; i < CLI_MAX; i++){
        allstate[strlen("ALLSTATE") + i] = ':';
        list[strlen("LIST") + i] = ':';
        touroku[i] = 0;
        s[i] = UNUSED;        
        pd[i] = UNUSED;
    }
    allstate[strlen("ALLSTATE") + i] = ';';
    allstate[strlen("ALLSTATE") + i + 1] = '\0';
    list[strlen("LIST") + i] = ';';
    list[strlen("LIST") + i + 1] = '\0';
    touroku[i] = 0;
    s[i] = UNUSED;
    pd[i] = UNUSED;
    for(i = 0; i < WIDTH; i++){
        puyo[strlen("PUYO:") + i] = '*';
    }
    puyo[strlen("PUYO:") + i] = ':';
    puyo[strlen("PUYO:") + i + 1] = '*';
    puyo[strlen("PUYO:") + i + 2] = ';';
    puyo[strlen("PUYO:") + i + 3] = '\0';



    /* $B@\B3<u$1IU$1MQ%=%1%C%H$r@8@.!#$3$N%=%1%C%H$O(BINET$B%I%a%$%s$G%9%H%j!<%`7?(B*/
    if ((newcomer = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("socket");
	exit(1);
    }

    setsockopt(newcomer, SOL_SOCKET, SO_REUSEADDR, (char *)&sockflag, sizeof(int));
    setsockopt(newcomer, IPPROTO_TCP, TCP_NODELAY, (char *)&sockflag, sizeof(int));

    /* bind()$B$G$N%(%i!<$rKI$0$?$a!"(Bsaddr$B$NCf?H$r(B0$B$K$9$k(B*/
    memset((char *)&saddr, 0, sizeof(saddr));

    /* bind()$B$G%=%1%C%H(Bs$B$N%"%I%l%9$r%P%$%s%I$9$k(B*/
    saddr.sin_family = AF_INET;
    saddr.sin_addr.s_addr = INADDR_ANY;
    saddr.sin_port = htons(PORT); /* $BDj?t(BPORT$B$O(B#define$B$7$?$b$N(B */

    if ((bind(newcomer, (struct sockaddr *)&saddr, sizeof(saddr))) == -1) {
	perror("bind");
	exit(1);
    }

    /* listen()$B$G@\B3$5$l$k$N$rBT$D(B*/
    if ((listen(newcomer, 5)) == -1) {
	perror("listen");
	exit(1);
    } 

    /* $B%a%$%s%k!<%W(B */
    while (1) {
	FD_ZERO(&readfds);    /* fds$B$N=i4|2=(B */
     	/* select()$B$N8!::BP>]$r@_Dj(B*/
        FD_SET(newcomer, &readfds);
	for (i = 0; i < max; i++) {
	    if (s[i] != UNUSED) {
		FD_SET(s[i], &readfds);
	    }
	}
        
	/* $B%a%C%;!<%8$,E~Ce$7$F$$$k%=%1%C%H$,$J$$$+D4$Y$k(B*/


	timeout.tv_sec = SELECT_TIMEOUT;
	timeout.tv_usec = 0;  

	                     /* katteni tuika shimashita!!!
				By kaidempa > fusa */




	if ((r = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)) == -1) {
	    perror("select");
	    exit(1);
	}
       
	/* $B%a%C%;!<%8$,$"$l$P!"MW5a$K1~$8$?=hM}$r9T$J$&(B*/
	for (i = 0; i < max; i++) {
	    if (s[i] != UNUSED) {
		if (FD_ISSET(s[i], &readfds)) {
		  /*s[i]$B$N%S%C%H$,N)$C$F$$$l$P!"(Bs[i]$B$K%a%C%;!<%8$,E~Ce$7$F$$$k(B*/
    
		    if ((msglen = read(s[i], str, sizeof(str))) == -1) {
			/* $B<u?.<:GT(B */
			perror("read");
		    } else if (msglen != 0) {  /* $B%a%C%;!<%8$N<u?.$K@.8y(B */
                        /*$BL?Na$N<oN`$r2r@O$7!"$=$l$K1~$8$?=hM}$r9T$J$&(B*/
                        str[msglen] = '\0';
               
			switch( analysis(str, state)){
                            case 1:    if(nickname_search(str) == 0){
                                           list_update(str, pd[i]);
                    
                                           for(j = 0; j < BEG_LINE; j++){
                                               puyo_make(pd[i]);
                                               write(s[i], puyo, strlen(puyo));          
					       /*
					       sleep(1);
					       */


                                           }
                                           touroku[i] = 1;
                                           broadcast(list, s, max);
				       }else{
                                
                                           write(s[i],"NICK:*;",strlen("NICK:*;"));
                                            close(s[i]);
                                            for(j =i ; j < CLI_MAX; j++){
                                                s[i] = UNUSED;
                                                pd[i]  = UNUSED;
                                                touroku[i] = 0;
					    }         
                                            max--;      
   			               }
                                       break;
                             
                            case 2:  
                                      connect_ = max;
                                      for(i = 0;i < connect_; i++){
                                          if(touroku[i] == 0){
                                              close(s[i]);
                                              for(j =i ; j < CLI_MAX; j++){
                                                   s[i] = UNUSED;
                                                   pd[i]  = UNUSED;
                                                   touroku[i] = 0;
					       }
                                               max--;
					  }
					}
                                       for(j = 0; j < max; j++){
                                           mystate[8] = '0' + pd[j];
                                           mystate[10] = '0' + 7;
                                           allstate_update(mystate);
				       }
                                      broadcast("START;", s, max);
                                      state = 1;
                                      time(&pre_time);
                                      break;

                            case 3:   pd0 = str[8] - '0';
                                      pd1 = str[10] - '0';
                                      n = str[12] -'0';
                                      for(j = 0; j < n; j++){
                                           puyo_make(pd0);
                                           write(s[pd1], puyo, strlen(puyo));
                                       }
                                      break;
                            case 4:  
                                      allstate_update(str);
                                     
                                
                                      if( (win_pd = winner(max)) != -1){
                                          win_make(str, win_pd);
                                          broadcast(str, s, max);
                                          
                                          state = 0;
                                          max = 0;
                                          for(i = 0; i < CLI_MAX; i++){
                                              allstate[strlen("ALLSTATE") + i] = ':';
                                              list[strlen("LIST") + i] = ':';
                                              touroku[i] = 0;
                                              s[i] = UNUSED;
                                              pd[i] = UNUSED;        
                                          }
                                          allstate[strlen("ALLSTATE") + i] = ';';
                                          allstate[strlen("ALLSTATE") + i + 1] = '\0';
                                          list[strlen("LIST") + i] = ';';
                                          list[strlen("LIST") + i + 1] = '\0';
                                          touroku[i] = 0;
                                          s[i] = UNUSED;
                                          pd[i] = UNUSED;
                                          for(i = 0; i < WIDTH; i++){
                                              puyo[strlen("PUYO:") + i] = '*';
                                          }
                                          puyo[strlen("PUYO:") + i ] = ':';
                                          puyo[strlen("PUYO:") + i + 1] = '*';
                                          puyo[strlen("PUYO:") + i + 2] = ';';
                                          puyo[strlen("PUYO:") + i + 3] = '\0';
                                          
                                          for(i = 0; i < max; i++){
                                              if( s[i] != UNUSED ){
                                                  close(s[i]);
                                              }
                                          }
                                      }
                                      break;
                    

                         }
                       
		    } else if( state == 0){
                      
                        nickname_erase(pd[i]);
                        close(s[i]);
                        for(j = i; j < max; j++){
                            s[j] = s[j+1];
                            touroku[j] = touroku[j + 1];
			}
                        max--;
		    }else{
		     
			close(s[i]);
			s[i] = UNUSED;
                        
                        mystate[8] = '0' + pd[i];
                        mystate[10] = '*';
                        allstate_update(mystate);
		    }
		}
	    }
	}	
	/* $B?7$?$J@\B3MW5a$KBP$9$k=hM}$r9T$&(B*/
	if (FD_ISSET(newcomer, &readfds) != 0) {
	   
	    /* $B@\B3$r3NN)$9$k(B accept() */
	    len = sizeof(caddr);
	    s[max] = accept(newcomer, (struct sockaddr *)&caddr, &len);
      
	    if (s[max] == -1) {
		perror(NULL);
		exit(1);
	    }
	    if (max <= CLI_MAX && state == 0) {
	
                pd[max] = max;
                touroku[max] = 1;
		max++;
	    } else {
     
		strcpy(str, "Server is too busy.\n");
		write(s[max], str, strlen(str));
		close(s[max]);
	    }
	}

        if( state == 1 ){
            time(&now_time);
            if(( now_time - pre_time) >= 5){
                broadcast(allstate,s,max);
                time(&pre_time);
         
	      }
	  }

    }
}


int get_port(char *argv[])
{

    int i, j;
    int number[10];
    int port = 0;
    int keta;

    if(strcmp(argv[1], "-p") != 0){
        return -1;
    }
  
    for(i = 0; i < strlen(argv[2]); i++){
        if(isdigit((argv[2])[i]) == 0){
            return -1;
	}
        number[i] = (argv[2])[i] - '0';
        keta = 1;
        for(j = i; j < strlen(argv[2]) - 1; j++){
            keta *= 10;
	}
        port += number[i] * keta;
    }

    return port;
}




/*$B@\B3%/%i%$%"%s%H$9$Y$F$KJ8;zNs(Bstr[]$B$rAw$k4X?t(B*/
void broadcast(char *str, int *s, int max)
{
    int i;
    for(i = 0; i < max; i++){
        if(s[i] != UNUSED){
   
            write(s[i],str,strlen(str));
  
        }
    }
}


/*$B%/%i%$%"%s%H$+$i$N%a%C%;!<%8(Bstr[]$B$H%2!<%`$N%b!<%I$rI=$9(Bstate*
 *$B$r<u$1<h$j!"L?Na$N<oN`$rI=$9?t;z$rJV$94X?t(B                  */
int analysis(char *str, int state)
{  
    if(state == 0){ /*$BEPO?<u$1IU$1Cf(B*/ 
        if(strncmp(str, "NICK:", strlen("NICK:")) == 0){ 
            return NICK;
        }else if(strncmp(str, "START;", strlen("START:")) == 0){ 
            return START;
        }else{ 
            return ERROR;
        }
    }else{  /*$B%2!<%`Cf(B*/
        if(strncmp(str, "REQUEST:", strlen("REQUEST:")) == 0){ 
            return REQUEST;
        }else if(strncmp(str, "MYSTATE:", strlen("MYSTATE:")) == 0){ 
            return MYSTATE;
        }else{ /*$B%(%i!<$J$i(B-1$B$rJV$9(B*/
            return ERROR;
        }
    }
}



int winner(int max)
{
    int count = 0;
    int winner = 0;
    int pd = 0;
    int i;

    for(i = strlen("ALLSTATE:"); allstate[i] != ';'; i++){
        switch(allstate[i]){
            case ':':  pd++;
                       break;
	    case '*':  count++;
                       break;
	    default:   winner = pd;
	}
    }

    if(count == (max - 1)){
        return winner;
    }else{
        return (-1);
    }
}



/*NICK$BL?Na$r$b$i$C$?>l9g$N=hM}!"6qBNE*$K$O!"(BNICK:$B$N8e$m$K=q$+$l$F$"$k(Bnickname*
 *$B$r<h$j=P$7!"(B"LIST:nickname0: . . . :nickname9;"$B$rJQ99$9$k4X?t(B              */

void list_update(char *str, int pd)
{
    int i, zure, space, length;
    int pointer1, pointer2;
    char nickname[10];

    for(i = 0; (nickname[i] = str[strlen("NICK:") + i]) != ';'; i++)
        ;
    nickname[i] = '\0';
    
    for(pointer1 = strlen("LIST:") , i = 0; i < pd; pointer1++){
        if(list[pointer1] == ':'){
            i++;
	}
    }
    pointer1--;
    for(pointer2 = pointer1 + 1; list[pointer2] != ':' && list[pointer2] != ';'; pointer2++)
        ;
    length = strlen(list);
    space = pointer2 - pointer1 - 1;
    if( space > strlen(nickname)){
        zure = space - strlen(nickname);
        for(i = pointer2; i <= length; i++){
            list[i - zure] = list[i];
        }
    }else if(space < strlen(nickname)){
        zure = strlen(nickname) - space;
        for(i = length; i >= pointer2; i--){
            list[i + zure] = list[i];
	}
    }

    for(i = 0; i < strlen(nickname) ; i++){
        list[pointer1 + i + 1] = nickname[i];
    }
                 
}





void nickname_erase(int pd)
{
    int i, space, length;
    int pointer1, pointer2;

    for(pointer1 = strlen("LIST:") , i = 0; i < pd; pointer1++){
        if(list[pointer1] == ':'){
            i++;
	}
    }
    pointer1--;
    for(pointer2 = pointer1 + 1; list[pointer2] != ':' && list[pointer2] != ';'; pointer2++)
        ;
    
    space = pointer2 - pointer1 - 1;
    length = strlen(list);
    list[length] = '\0';
    for(i = pointer2; i <= length ; i++){
        list[i - space] = list[i];
    }
    for(i = pointer1 + 1; (list[i - 1] = list[i]) != '\0'; i++)
        ;
    
    list[strlen(list) - 1] = ':';
    list[strlen(list)] = ';';
    list[strlen(list) + 1] = '\0';                 
}



void allstate_update(char *str)
{
    int i, j, count, zure, space, length;
    int pointer1, pointer2;
    char pd[10], yabasa[10];

    for(i = 0; (pd[i] = str[strlen("MYSTATE:") + i]) != ':'; i++)
        ;
    pd[i] = '\0';
    for(j = 0, i++; (yabasa[j] = str[strlen("MYSTATE:") + i]) != ';' ; j++,i++)
        ;
    yabasa[j] = '\0';
    count = pd[0] - '0';
    for(pointer1 = strlen("ALLSTATE:") , i = 0; i < count; pointer1++){
        if(allstate[pointer1] == ':'){
            i++;
	}
    }
    pointer1--;
    for(pointer2 = pointer1 + 1; allstate[pointer2] != ':' && allstate[pointer2] != ';'; pointer2++)
        ;
    length = strlen(allstate);
    space = pointer2 - pointer1 - 1;
    if( space > strlen(yabasa)){
        zure = space - strlen(yabasa);
        for(i = pointer2; i <= length; i++){
            allstate[i - zure] = allstate[i];
        }
    }else if(space < strlen(yabasa)){
        zure = strlen(yabasa) - space;
        for(i = length; i >= pointer2; i--){
            allstate[i + zure] = allstate[i];
	}
    }

    for(i = 0; i < strlen(yabasa) ; i++){
        allstate[pointer1 + i + 1] = yabasa[i];
    }
                 
}










void win_make(char *str, int pd)
{
    str[0] = 'W';
    str[1] = 'I';
    str[2] = 'N';
    str[3] = ':';
    str[4] = (pd + '0');
    str[5] = ';';
    str[6] = '\0';
}

/*$B$*<YKb$W$h$NAw$j<j$rI=$9%W%l%$%d%G%9%/%j%W%?(Bpd$B$r<u$1<h$j!"(B*
 *$BJ8;zNs(Bpuyo[]$B$rA`:n$7!"(BPUYO$BL?Na$r:n@.$9$k4X?t(B              */
void puyo_make(int pd)
{
    int  i;
    int  coulor(void);
    puyo[strlen("PUYO:") + WIDTH + 1] = (pd + '0');
    for(i = 0; i < WIDTH; i++){
            puyo[strlen("PUYO:") + i] = (coulor() + '0');
    }
}


/*1,2,3,4$B$N$$$:$l$+$r%i%s%@%`$KH/@8$5$;$k4X?t(B*/
int coulor(void)
{
    double rand;
    float  randomi(void);

    rand = randomi(); 
    if(rand < 0.25){      /*0.25$BL$K~$J$i(B1$B$rJV$9(B*/
        return 1;
    }else if(rand < 0.5){  /*0.25$B0J>e(B0.5$BL$K~$J$i(B2$B$rJV$9(B*/
        return 2;
    }else if(rand < 0.75){ /*0.5$B0J>e(B0.75$BL$K~$J$i(B3$B$rJV$9(B*/
        return 3;
    }else{                   /*0.75$B0J>e$J$i(B4$B$rJV$9(B*/
        return 4;
    }
}




int nickname_search(char *str)
{
    int i, space;
    int pointer1, pointer2;
    int same = 0;
    char nickname[15];
    char usedname[15];

    for(i = 0; (nickname[i] = str[strlen("NICK:") + i]) != ';'; i++)
        ;
    nickname[i] = '\0';
     
    for(pointer1 = strlen("LIST"); list[pointer1] != ';'; pointer1 = pointer2){
        for(pointer2 = pointer1 + 1; list[pointer2] != ':' && list[pointer2] != ';'; pointer2++)
            ;
        space = pointer2 - pointer1 - 1; 
        if( space > 0){   
            for(i = 0; i < space ; i++){
                usedname[i] = list[pointer1 + i + 1];
            }
            usedname[i] = '\0';

            if(strcmp(nickname,usedname) == 0){
                same = 1;
	    }
        }
     }
     return same;
    
}

            

    
/*0$B0J>e(B1$BL$K~$NMp?t$rH/@8$5$;$k4X?t(B*/
float randomi(void)
{
    static unsigned long x = RANDOM_INIT_VALUE;

    x = LAMBDA * x + RANCONST;
    return (float) x / RAN2_32;
}














